#! /usr/bin/python
"""
!!!! this version supports multi-processing which split the funSP into different sub-processes for full sampling to save memory
!!!! MEAGA requires networkx (see ggscoreGraph.py)


input: 
i) marker2gene (ie map + genes seperated by ;)
ii) gene2fun
iii) interested markers
iv) directory for Fun SP path files (G1 G2 shortest_distance shortest_path)
v) output directory + prefix

optional input:
i) # of processes (default=1)
ii) # of random samplings (default=10000)
iii) min # of fun genes (default=5)
iv) max # of fun genes (default=1000)
v) min # of interested genes in fun (default=3)
vi) D_max: distance used for estimating distance between isolated node with other connected components

output:
i) MEAGA result : Fun \t #FunGenes \t #IntFun genes \t # Connected Graph \t #Int_Fun_genes/#Connected_Graph \t S_T\t p-val \t adj pval


ii) MEAGA (funichSPtree) trees: Fun \t  gene1 \t gene2 (NA if gene1 is isolated) \t weight

Algorithm:
(i) obtain sig Genes from sig markers --> G
(ii) for every fun, get overlap genes with G --> G'
(iii) using G' to get subgraph(s) from the function's SP file (including SP length and SP (intermediate nodes in SP not necessarily in G'))--> H
(iv) for each connected component in H, identify MST
(v) perform sampling and re-do (i-iv)
"""

from optparse import OptionParser
import sys
import os
#sys.path.append("/net/fantasia/home/alextsoi/code/Network")
sys.path.append(os.path.dirname(os.path.realpath(sys.argv[0])))
from ggscoreGraph import *
from FunMST import *
from computeST import *
from random import *
from threading import *
from multiprocessing import *
from math import *

def main():
	usage="usage: %prog -s marker2gene -g gene2fun -i intmarkers -d funSPdir -o outDir [-t numProcess -n numsamples -m minFun -M maxFun -a minIntFun]"
	parser=OptionParser(usage)
	parser.add_option("-s", "--marker2gene",dest="marker2genefile",help="file name of marker2gene file (map file + associated genes delimited by\";\")")
	parser.add_option("-g", "--gene2fun",dest="gene2funfile",help="file name of annotation file (gene2anno2source)")
	parser.add_option("-i", "--intmarkers",dest="intmarkerfile",help="file name of the interested markers")
	parser.add_option("-d", "--funSPdir",dest="funSPdir",help="directory storing the prefix folders for the function SP files)")
	parser.add_option("-o", "--outDir",dest="outDir",help="output directory and prefix")
	
	### default value for t, n, m, M, a, and D are 1, 10000, 5, 1000, and 3, 12??
	parser.add_option("-t","--numProcess",dest="numProcess",type="int",default=1)
	parser.add_option("-n","--numsamples",dest="numsamples",type="int",default=10000)
	parser.add_option("-m","--minFun",dest="minFun",type="int",default=5,help="minimum # of genes in function")
	parser.add_option("-M","--maxFun",dest="maxFun",type="int",default=1000,help="maximum # of genes in function")
	parser.add_option("-a","--minIntFun",dest="minIntFun",type="int",default=3)
	parser.add_option("-D","--Dmax",dest="Dmax",type="int",default=12)

	(options,args)=parser.parse_args()
	if options.marker2genefile==None or options.gene2funfile==None or options.intmarkerfile==None or options.funSPdir==None or options.outDir==None:
		parser.error("Input arguments are not fully specified")

	numProcess=int(options.numProcess)
	numsamples=int(options.numsamples)
	minFun=int(options.minFun)
	maxFun=int(options.maxFun)
	minIntFun=int(options.minIntFun)
	Dmax=int(options.Dmax)

	############################################################
	### parsing marker2gene file
	marker2gene={}
	count=0
	for tempstr in open(options.marker2genefile):
		tempstr=tempstr.rstrip().split()
		marker2gene[tempstr[1]]=tempstr[4].split(";")
		count+=1
	print "Processed "+str(count)+" markers in marker2gene file..."

	### parsing gene2fun file
	fun2gene={}
	count=0
	for tempstr in open(options.gene2funfile):
		count+=1
		tempstr=tempstr.rstrip().split()
		if tempstr[1] not in fun2gene:
			fun2gene[tempstr[1]]=[tempstr[0]]
		else:
			fun2gene[tempstr[1]].append(tempstr[0])
	print "Processed "+str(count)+" gene2fun records..."

	### parsing intmarker file
	intmarkers=[]
	count=0
	for tempstr in open(options.intmarkerfile):
		count+=1
		tempstr=tempstr.rstrip().split()
		intmarkers.append(tempstr[0])
	print "Processed "+str(count)+" interested markers..."

	### get the interested genes from intmarkers:
	intgenes=[]
	for i in intmarkers:
		intgenes=intgenes+marker2gene[i]
	
	intgenes=list(set(intgenes))
	# remove NA intgenes
	intgenes=list(set(intgenes)-set(["NA"]))
	print "There are totally "+str(len(intgenes))+" interested genes from the interested markers...."

#############################################################################################################
	### get those valid functions
	intfuns=[]
	for i in fun2gene:
		tempfungenes=fun2gene[i]
		if len(tempfungenes)<minFun or len(tempfungenes)>maxFun:
			continue
		tempfunintgenes=list(set(tempfungenes)&set(intgenes))
		if len(tempfunintgenes)>=minIntFun:
			intfuns.append(i)
	print "There are totally "+str(len(intfuns))+" interested/valid functions we are examining..."

###########################################################################################################
	### for those filtered functions, get their funSP files
	funSP={}
	print "Loading SP graphs for interested/valid functions..."
	for i in intfuns:
		prefix=i[0]
		tempGraph=Graph()
		for tempstr in open(options.funSPdir+prefix+"/"+i):
			if tempstr[0]=="#":
				continue
			tempstr=tempstr.rstrip().split()
			tempGraph.add_edge(tempstr[0],tempstr[1],weight=tempstr[2],sp=tempstr[3].split(","))
		funSP[i]=tempGraph
		
	print "finished loading."
	##############################################################
	### compute S_T for interested genes in interested functions
	### could not use thread_funslist_MST_ST as we need to keep track of tempFunMST here
	funichSPtreeoutf=open(options.outDir+"MEAGAtree",'w')	
	numfungenes=[]
	numfunintgenes=[]
	numconnectedgraphs=[]
	intfun_ST=[]
	avgintfungenes_perconnectedgraph=[]

	for i in intfuns:	
		tempfunSP=Graph(funSP[i])
		tempfungenes=fun2gene[i]
		tempfunintgenes=list(set(tempfungenes)&set(intgenes))
		tempFunMST=FunMST(tempfunSP,tempfunintgenes)
		### compute ST
		tempST=computeST(tempFunMST,tempfunintgenes,Dmax)
		intfun_ST.append(str(tempST))
		numfungenes.append(str(len(tempfungenes)))
		numfunintgenes.append(str(len(tempfunintgenes)))
		numconnectedgraphs.append(str(len(connected_components(tempFunMST))))
		avgintfungenes_perconnectedgraph.append(str(float(len(tempfunintgenes))/len(connected_components(tempFunMST))))
		### output funichSPtree
		for j in connected_components(tempFunMST):
			if len(j)==1:
				funichSPtreeoutf.write(i+"\t"+j[0]+"\tNA\tNA\n")
			else:
				tempFunMST_sub=tempFunMST.subgraph(j)
				for k in tempFunMST_sub.edges(data=True):
					funichSPtreeoutf.write(i+"\t"+k[0]+"\t"+k[1]+"\t"+str(k[2]['weight'])+"\n")
	funichSPtreeoutf.close()
	#############################################################
	### in order for every subprocess to have the same set of samples genes, we need to pre-sample random genes !!!
	print "Sampling genes...."
	randgeneslist=[]
	for i in range(numsamples):
		randgeneslist.append(sampleGenesfromMarkers(marker2gene,intgenes))

	##############################################################################
	### compute S_T for random genes in interested functions
	print "Compute S_t for random sampled genes...."
	### random_ST is a dict (interested functions) of lists (samples) of list (intfun_ST, #funintgenes, avgfunintgenes/connectedgraph)
	random_ST={}
	processes=[]
	queues=Queue()

	avgnumintfuns=int(ceil(len(intfuns)/float(numProcess)))
	culnumintfuns=0
	for i in range(numProcess):
		if i==(numProcess-1):
			tempintfuns=intfuns[culnumintfuns:]
		else:
			tempintfuns=intfuns[culnumintfuns:(culnumintfuns+avgnumintfuns)]
		culnumintfuns+=avgnumintfuns
		tempfunSP=dict([(a,funSP[a]) for a in tempintfuns])

		processes.append(Process(target=process_funslist_MST_ST,args=(tempfunSP, tempintfuns, fun2gene,Dmax,numsamples,queues,randgeneslist)))

	### start new processes
	for t in processes:
		t.start()
	
	### gather the random_ST
	for t in range(numProcess):
		random_ST.update(queues.get())

	################################################################
	### print out S_T and S_t
	
	S_Toutf=open(options.outDir+"intgenes_S_T",'w')
	S_toutf=open(options.outDir+"randomgenes_S_t",'w')
	
	numfunintgenes_outf=open(options.outDir+"numfunintgenes",'w')
	numfunintgenes_sample_outf=open(options.outDir+"numfunintgenes_sample",'w')
	
	avgintfungenes_perconnectedgraph_outf=open(options.outDir+"avgintfungenes_perconnectedgraph",'w')
	avgintfungenes_perconnectedgraph_sample_outf=open(options.outDir+"avgintfungenes_perconnectedgraph_sample",'w')

	S_Toutf.write('\t'.join(intfuns)+"\n")
	S_Toutf.write('\t'.join(intfun_ST)+"\n")

	numfunintgenes_outf.write('\t'.join(intfuns)+"\n")
	numfunintgenes_outf.write('\t'.join(numfunintgenes)+"\n")

	avgintfungenes_perconnectedgraph_outf.write('\t'.join(intfuns)+"\n")
	avgintfungenes_perconnectedgraph_outf.write('\t'.join(avgintfungenes_perconnectedgraph)+"\n")

	S_toutf.write('\t'.join(intfuns)+"\n")
	numfunintgenes_sample_outf.write('\t'.join(intfuns)+"\n")
	avgintfungenes_perconnectedgraph_sample_outf.write('\t'.join(intfuns)+"\n")


	for i in range(numsamples):
		for f in range(len(intfuns)):
			tempresult=random_ST[intfuns[f]][i]
			if f==0:
				S_toutf.write(tempresult[0])
				numfunintgenes_sample_outf.write(tempresult[1])
				avgintfungenes_perconnectedgraph_sample_outf.write(tempresult[2])
			else:
				S_toutf.write("\t"+tempresult[0])
				numfunintgenes_sample_outf.write("\t"+tempresult[1])
				avgintfungenes_perconnectedgraph_sample_outf.write("\t"+tempresult[2])
		S_toutf.write("\n")
		numfunintgenes_sample_outf.write("\n")
		avgintfungenes_perconnectedgraph_sample_outf.write("\n")

	S_Toutf.close()
	S_toutf.close()
	
	numfunintgenes_outf.close()
	numfunintgenes_sample_outf.close()
	
	avgintfungenes_perconnectedgraph_outf.close()
	avgintfungenes_perconnectedgraph_sample_outf.close()

	######################################################
	### compute p-valule and adjusted p-value

	os.system("Rscript --vanilla "+ os.path.dirname(os.path.realpath(sys.argv[0]))+"/computeP_funichSP.R "+options.outDir+"intgenes_S_T "+options.outDir+"randomgenes_S_t "+options.outDir+"numfunintgenes "+options.outDir+"numfunintgenes_sample "+options.outDir+"avgintfungenes_perconnectedgraph "+options.outDir+"avgintfungenes_perconnectedgraph_sample "+options.outDir+".tempPvals")
	funs2p={}
	for tempstr in open(options.outDir+".tempPvals"):
		tempstr=tempstr.rstrip().split()
		funs2p[tempstr[0]]=[tempstr[1],tempstr[2],tempstr[3],tempstr[4],tempstr[5],tempstr[6]]

	############################################################################################################
	### outoutput all results
	#funichresultoutf=open(options.outDir+"MEAGAresult",'w')
	#funichresultoutf.write("Funs\tNumFunGenes\tNumFunIntGenes\tNumConnectedGraphs\tAvgIntFunGenes_perConnectedGraphs\tS_T\tpval\tadj-pval\tALI_pval\tALI_adjpval\tAvgGeneTree_pval\tAvgGeneTree_adjpval\n")
	#for i in range(len(intfuns)):
	#	funichresultoutf.write(intfuns[i]+"\t"+str(numfungenes[i])+"\t"+str(numfunintgenes[i])+"\t"+str(numconnectedgraphs[i])+"\t"+str(avgintfungenes_perconnectedgraph[i])+"\t"+str(intfun_ST[i])+"\t"+'\t'.join(funs2p[intfuns[i]])+"\n")
	
	#funichresultoutf.close()


	############################################################################################################
	### output only MEAGA result
	funichresultoutf=open(options.outDir+"MEAGAresult",'w')
        funichresultoutf.write("Funs\tNumFunGenes\tNumFunIntGenes\tNumConnectedGraphs\tAvgIntFunGenes_perConnectedGraphs\tS_T\tpval\tadj-pval\n")
        for i in range(len(intfuns)):
        	funichresultoutf.write(intfuns[i]+"\t"+str(numfungenes[i])+"\t"+str(numfunintgenes[i])+"\t"+str(numconnectedgraphs[i])+"\t"+str(avgintfungenes_perconnectedgraph[i])+"\t"+str(intfun_ST[i])+"\t"+funs2p[intfuns[i]][0]+"\t"+funs2p[intfuns[i]][1]+"\n")

        funichresultoutf.close()


	#######################################################
	### remove those intermediate files:
	tempstring="rm -f "+options.outDir+".tempPvals"
	os.system(tempstring)

	tempstring="rm -f "+options.outDir+"intgenes_S_T"
        os.system(tempstring)

	tempstring="rm -f "+options.outDir+"randomgenes_S_t"
        os.system(tempstring)

	tempstring="rm -f "+options.outDir+"numfunintgenes"
        os.system(tempstring)

	tempstring="rm -f "+options.outDir+"numfunintgenes_sample"
        os.system(tempstring)

	tempstring="rm -f "+options.outDir+"avgintfungenes_perconnectedgraph"
        os.system(tempstring)

	tempstring="rm -f "+options.outDir+"avgintfungenes_perconnectedgraph_sample"
        os.system(tempstring)

#####################################################################
def sampleGenesfromMarkers(marker2gene,intgenes):
	### from marker2gene, sample marker until the # of sampled genes equal the # of intgenes
	rgenes=[]
	markers=marker2gene.keys()
	while (len(rgenes)<len(intgenes)):
		randm = markers[randint(0,len(markers)-1)]
		randg = marker2gene[randm]
		### the sampled genes could not be intgenes
		###randg = list(set(randg)-set(intgenes))
		randg = list(set(randg)-set(["NA"]))
		rgenes+=randg
	if (len(rgenes)>len(intgenes)):
		rgenes=rgenes[0:len(intgenes)]
	return rgenes
	

#####################################################################
def process_funslist_MST_ST(funSP, intfunslist, fun2gene,Dmax, numsamples,que,randgenelist):
	### given a list of interested functions, each goes through the FunMST and computeST and output the list of ST values for each interested functions
	### intfun_ST is a list storing lists of 3 elements: intfun_ST, #funintgenes, avgfunintgenes/connectedgraph
	intfun_ST={}
	count=0
	for i in intfunslist:
		count+=1
		if count%50==0:
			print "Processed "+str(count)+" intfunslist in one of the processes.."
		tempfunSP=Graph(funSP[i])
		tempfungenes=fun2gene[i]
		samples_intfun_ST=[]
		for s in range(numsamples):
			randgenes=randgenelist[s]
			tempfunintgenes=list(set(tempfungenes)&set(randgenes))
			tempFunMST=FunMST(tempfunSP,tempfunintgenes)
			### compute ST
			tempST=computeST(tempFunMST,tempfunintgenes,Dmax)
			### !!! if len(tempfunintgenes) equals zero, set (float(len(tempfunintgenes))/len(connected_components(tempFunMST))) equals zero
			if len(tempfunintgenes)==0:
				samples_intfun_ST.append([str(tempST),str(0),str(0)])
			else:	
				samples_intfun_ST.append([str(tempST),str(len(tempfunintgenes)),str(float(len(tempfunintgenes))/len(connected_components(tempFunMST)))])
		intfun_ST[i]=samples_intfun_ST
	que.put(intfun_ST)


###################################################################
if __name__ == "__main__":
	main()

